from dataclasses import dataclass

from numberize import Numberizer
from numberize.map_text import ReplacedNumeral


@dataclass
class QuestionMapAnswer:
    text: str
    map: list
    out: str


def test_replace_by_map():
    data_set = [
        QuestionMapAnswer(
            'аодопддп', [ReplacedNumeral('_', 0, 5)], '_ддп'
        ),
        QuestionMapAnswer(
            'аодопддп',
            [ReplacedNumeral('_', 0, 5), ReplacedNumeral(5, 6, 7)],
            '_д5п'
        ),
    ]
    numba = Numberizer('ru')
    for item in data_set:
        ans = numba._replace_by_map(item.map, item.text)
        assert ans == item.out, ans


def test_replace_numerals_ru():
    text_replaced = {
        'девять-восемь, тремя миллионами, шестьсот тысяч и трёх людей':
        '9-8, 3000000, 600000 и 3 людей',
        '"сто" - слово очень простое, состоит из трёх букв':
        '"100" - слово очень простое, состоит из 3 букв',

        'двадцать пять': '25',
        'двадцать пять и семь': '25 и 7',
        'двадцать пять тысяч и Вася': '25000 и Вася'
    }
    numba = Numberizer('ru')
    for text in text_replaced:
        ans = numba.replace_numerals(text)
        assert ans == text_replaced[text], ans


def test_replace_numerals_uk():
    text_replaced = {
        """дев'ять-вісім, трьома мільйонами, шістсот тисяч та трьох людей""":
        '9-8, 3000000, 600000 та 3 людей',
        """
        "сто" - слово наївне, складається з трьох букв'
        """:
        """
        "100" - слово наївне, складається з 3 букв'
        """,

        """двадцять п'ять""": """25""",
        """двадцять п'ять та сім""": '25 та 7',
        """двадцять п'ять тисяч та Василь""": '25000 та Василь'
    }
    numba = Numberizer('uk')
    for text in text_replaced:
        ans = numba.replace_numerals(text)
        assert ans == text_replaced[text], ans


def test_replace_numerals_en():
    text_replaced = {
        """In American English, the conjunction "and" is generally not used before tens or ones: one hundred twenty-three (123); four hundred seven (407); three thousand five hundred thirty-eight (3,538); seventy-three thousand five (73,005); two million six hundred twenty-five thousand three hundred ten (2,625,310); five million three hundred thousand fifty (5,300,050).""":
        """In American English, the conjunction "and" is generally not used before tens or ones: 123 (123); 407 (407); 3538 (3,538); 73005 (73,005); 2625310 (2,625,310); 5300050 (5,300,050).""",

        """twenty-five""": '25',
        """twenty-seven and seven""": '27 and 7',
        """twenty-five thousand and Vasya""": '25000 and Vasya'
    }
    numba = Numberizer('en')
    for text in text_replaced:
        ans = numba.replace_numerals(text)
        assert ans == text_replaced[text], ans


test_replace_numerals_en()




millenniums = {
    'тысяча': 1e3,
    'миллион': 1e6,
    'миллиард': 1e9,
    'триллион': 1e12
}


all_num = {
    'тысяча': 1e3,
    'миллион': 1e6,
    'миллиард': 1e9,
    'триллион': 1e12,
    'сто': 100,
    'двести': 200,
    'триста': 300,
    'четыреста': 400,
    'пятьсот': 500,
    'шестьсот': 600,
    'семьсот': 700,
    'восемьсот': 800,
    'девятсот': 900,
    'двадцать': 20,
    'тридцать': 30,
    'сорок': 40,
    'пятьдесят': 50,
    'шестдесят': 60,
    'семьдесят': 70,
    'восемьдесят': 80,
    'девяносто': 90,
    'один': 1,
    'два': 2,
    'три': 3,
    'четыре': 4,
    'пять': 5,
    'шесть': 6,
    'семь': 7,
    'восемь': 8,
    'девять': 9,
    'десять': 10,
    'одинадцать': 11,
    'двенадцать': 12,
    'тринадцать': 13,
    'четырнадцать': 14,
    'пятнадцать': 15,
    'шестнадцать': 16,
    'семнадцать': 17,
    'восемнадцать': 18,
    'девятнадцать': 19
}



from nltk.tokenize import word_tokenize
from collections import namedtuple


from numberize.analyze import Analyzer, Checker

import numberize.dict_en as dict_en

ReplacedNumeral = namedtuple('ReplacedNumeral', ['number', 'start', 'end'])


def calculate_a_num(numbers) -> int:
    res, group = 0, 0
    for num in numbers:
        if num < 1E3:
            group += num
            continue
        if group != 0:
            res += group*num
            group = 0
            continue
        res += num
    else:
        res += group
    return int(res)


def calculate_a_num_eng(numbers) -> int:
    res, h_group, m_group = 0, 0, 0
    print(numbers)
    for num in numbers:
        if num == 100:
            if h_group != 0:
                m_group += h_group*num
                h_group = 0
                continue
            m_group += num
        if num in (1000, 1000000, 1000000000):
            if h_group and m_group:
                res += (h_group+m_group)*num
                h_group, m_group = 0, 0
            elif h_group:
                res += h_group*num
                h_group = 0
            elif m_group:
                res += m_group*num
                m_group = 0
            else:
                res += num
            continue
        h_group += num
    else:
        res += m_group + h_group
    return int(res)


def eng_numeral(tok: str) -> bool:
    if not tok.isalpha():
        if '-' in tok:
            two_parts = tok.split('-')
            if len(two_parts) != 2:
                return False
            if two_parts[0] in dict_en.all_num \
                    and two_parts[1] in dict_en.all_num:
                return True
        return False
    if tok in dict_en.all_num:
        return True


class Mapper:
    def __init__(self, analyzer: 'Analyzer', checker: 'Checker'):
        self._replacement_map = []
        self._current_word = ''
        self._current_numeral = []
        self._start_of_numeral, self._end_of_numeral = None, None

        self._morph = analyzer
        self._check = checker

    def _clear(self):
        self._replacement_map = []
        self._current_word = ''
        self._current_numeral = []
        self._start_of_numeral, self._end_of_numeral = None, None

    def _update_word(self, char: str, start: int) -> None:
        if not self._current_word:
            self._start_of_numeral = start
        self._current_word += char

    def _update_replacement_map(self) -> None:
        if self._current_numeral:
            num_start = self._current_numeral[0].start
            num_end = self._current_numeral[-1].end
            num = calculate_a_num(
                (x.number for x in self._current_numeral)
            )
            self._replacement_map.append(
                ReplacedNumeral(num, num_start, num_end)
            )
            self._current_numeral = []

    def _update_numeral(self, end: int) -> None:
        self._end_of_numeral = end
        parsed = self._check.get_parsed(self._morph.parse(self._current_word))
        if parsed:
            self._current_numeral.append(
                ReplacedNumeral(
                    self._check.get_num(parsed.normal_form),
                    self._start_of_numeral,
                    self._end_of_numeral
                )
            )
        else:
            self._update_replacement_map()
        self._current_word = ''

    @staticmethod
    def _eng_replacement_map(text: str) -> list:
        numeral, replacement_map = [], []
        for i, tok in enumerate(word_tokenize(text)):
            if eng_numeral(tok):
                if '-' not in tok:
                    numeral.append(
                        ReplacedNumeral(
                            dict_en.all_num[tok],
                            i,
                            i + 1
                        )
                    )
                    continue
                l, r = tuple(tok.split('-'))
                numeral.append(
                    ReplacedNumeral(
                        dict_en.all_num[l]+dict_en.all_num[r],
                        i,
                        i + 1
                    )
                )
                continue
            if numeral:
                replacement_map.append(
                    ReplacedNumeral(
                        calculate_a_num_eng([x.number for x in numeral]),
                        numeral[0].start,
                        numeral[-1].end
                    )
                )
                numeral = []
        else:
            if numeral:
                replacement_map.append(
                    ReplacedNumeral(
                        calculate_a_num(x.number for x in numeral),
                        numeral[0].start,
                        numeral[-1].end
                    )
                )
        return replacement_map

    def get_replacement_map(self, text: str) -> list:
        self._clear()

        if hasattr(self._morph, '_lang'):
            return self._eng_replacement_map(text)

        for current_position, char in enumerate(text):
            if self._check.is_cyrillic(char):
                self._update_word(char, start=current_position)
            elif self._current_word:
                if char.isspace():
                    self._update_numeral(end=current_position)
                    continue
                self._update_numeral(end=current_position)
                self._update_replacement_map()
            elif not char.isspace():
                self._update_replacement_map()
        else:
            if self._current_word:
                self._update_numeral(end=len(text))
                self._update_replacement_map()
            if self._current_numeral:
                self._update_replacement_map()
        return self._replacement_map





from nltk.tokenize import word_tokenize
from nltk.tokenize.treebank import TreebankWordDetokenizer


import numberize.map_text as map_text

from numberize.analyze import Analyzer, Checker


class Numberizer:
    _languages = ('ru', 'uk', 'en')

    def __init__(self, lang: str = 'ru'):
        """
        :param lang: 'ru' - russian, 'uk' - ukrainian, 'en' - english
        """
        if lang in Numberizer._languages:
            self._analyzer = Analyzer(lang)
            self._checker = Checker(lang)
        else:
            raise Exception(
                f'{lang} is not supported language.\
                 Try one of these: {Numberizer._languages}'
            )

    @classmethod
    def supported_languages(cls):
        return cls._languages

    def _replace_by_map(self, replacement_map, text) -> str:
        if hasattr(self._analyzer, '_lang'):
            new_text = []
            text = word_tokenize(text)
            prev_end = 0
            for item in replacement_map:
                new_text += text[prev_end:item.start] + [str(item.number)]
                prev_end = item.end
            return TreebankWordDetokenizer().detokenize(
                new_text + text[prev_end:]
            )

        new_text = ''
        prev_end = 0
        for item in replacement_map:
            new_text += text[prev_end:item.start] + str(item.number)
            prev_end = item.end

        return new_text + text[prev_end:]

    def replace_numerals(self, text: str) -> str:
        mapper = map_text.Mapper(self._analyzer, self._checker)
        replacement_map = mapper.get_replacement_map(text)
        return self._replace_by_map(replacement_map, text)



from numberize.analyze import Checker, Analyzer


def test_is_cyrillic_ru():
    char_bool = {
        'f': False,
        '.': False,
        'А': True,
        'в': True,
        'ї': False
    }
    checker = Checker('ru')
    for char in char_bool:
        ans = checker.is_cyrillic(char)
        assert ans == char_bool[char], char


def test_is_cyrillic_uk():
    char_bool = {
        'f': False,
        '.': False,
        'А': True,
        'в': True,
        'ї': True,
        "'": True
    }
    checker = Checker('uk')
    for char in char_bool:
        ans = checker.is_cyrillic(char)
        assert ans == char_bool[char], char


def test_get_parsed_ru():
    words = ('один', "сто", 'десять', 'тысяча')
    checker = Checker('ru')
    analyzer = Analyzer('ru')
    for word in words:
        assert checker.get_parsed(analyzer.parse(word))


def test_get_parsed_uk():
    words = ("один", "сто", "тисяча", "п'яти")
    checker = Checker('uk')
    analyzer = Analyzer('uk')
    for word in words:
        assert checker.get_parsed(analyzer.parse(word))



millenniums = {
    "тисяча": 1E3,
    "мільйон": 1E6,
    "мільярд": 1E9,
    "трильйон": 1E12,
}

all_num = {
    "тисяча": 1E3,
    "мільйон": 1E6,
    "мільярд": 1E9,
    "трильйон": 1E12,
    "сто": 100,
    "двісті": 200,
    "триста": 300,
    "чотириста": 400,
    "п'ятсот": 500,
    "шістсот": 600,
    "сімсот": 700,
    "вісімсот": 800,
    "дев'ятсот": 900,
    "один": 1,
    "два": 2,
    "три": 3,
    "чотири": 4,
    "п'ять": 5,
    "шість": 6,
    "сім": 7,
    "вісім": 8,
    "дев'ять": 9,
    "десять": 10,
    "одинадцять": 11,
    "дванадцять": 12,
    "тринадцять": 13,
    "чотирнадцять": 14,
    "п'ятнадцять": 15,
    "шістнадцять": 16,
    "сімнадцять": 17,
    "вісімнадцять": 18,
    "дев'ятнадцять": 19,
    "двадцять": 20,
    "тридцять": 30,
    "сорок": 40,
    "п'ятдесят": 50,
    "шістдесят": 60,
    "сімдесят": 70,
    "вісімдесят": 80,
    "дев'яносто": 90,
}



import numberize

import numberize.analyze
import numberize.map_text


def test_calculate_a_num():
    numbers_res = {
        (300, 80, 1): 381,
        (3, 1E6): 3000000,
        (5,): 5,
        (300, 60, 1E3, 80, 3): 360083
    }
    for numbers in numbers_res:
        ans = numberize.map_text.calculate_a_num(numbers)
        assert ans == numbers_res[numbers], ans


def test_get_replacement_map_ru():
    text_map = {
        'двадцать пять': [(25, 0, 13)],
        'восьмидесяти пяти и семи': [(85, 0, 17), (7, 20, 24)],
        'восьмидесяти пяти и семи .ю.карандашей': [(85, 0, 17), (7, 20, 24)],
        'восемь-девять, тремя миллионами гривен и "восемьюстами тысячами" ':
        [(8, 0, 6), (9, 7, 13), (3000000, 15, 31), (800000, 42, 63)]
    }
    analyzer = numberize.analyze.Analyzer('ru')
    checker = numberize.analyze.Checker('ru')
    mapper = numberize.map_text.Mapper(analyzer, checker)
    for text in text_map:
        ans = mapper.get_replacement_map(text)
        assert ans == text_map[text], ans




import pymorphy2

from numberize.numeric_dict import nums


class Analyzer:
    def __init__(self, lang: str):
        if lang == 'en':
            self._lang = 'en'
        else:
            self._morph = pymorphy2.MorphAnalyzer(lang=lang)

    def _get_lang(self):
        return self._lang

    def parse(self, word: str):
        return self._morph.parse(word)


class Checker:
    def __init__(self, lang: str):
        self._lang = lang

    def is_cyrillic(self, char) -> bool:
        alphabet = {
            'ru': 'абвгдеэёжзийклмнопрстуфхцшчщюяыьъ',
            'uk': 'йцукенгшщзхїєждлорпавіфячсмитьбюґ\''
        }
        return char.lower() in alphabet[self._lang]

    def _ru_get_parsed(self, parsed: list):
        possible = parsed[0]
        part_of_speech, normal_form = possible.tag.POS, possible.normal_form
        if part_of_speech == 'NUMR' or normal_form == 'сто' \
            or normal_form == 'один' \
                or normal_form in nums[self._lang].millenniums:
            return possible

    def _uk_get_parsed(self, parsed: list):
        for possible in parsed:
            if possible.tag.POS == 'NUMR' \
                    or possible.normal_form in nums[self._lang].millenniums:
                return possible

    def get_parsed(self, parsed: list):
        languages = {
            'ru': self._ru_get_parsed,
            'uk': self._uk_get_parsed
        }
        return languages[self._lang](parsed)

    def get_num(self, normal_form: str):
        return nums[self._lang].all_num[normal_form]



import pytest


import numberize
from numberize.analyze import Analyzer


def test__init__():
    with pytest.raises(ValueError):
        Analyzer('ku')
    for lang in numberize.Numberizer.supported_languages():
        assert Analyzer(lang), lang




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="numberize",
    version="0.1.7a",
    author="YemchenkoDS",
    author_email="emchenko@dlit.dp.ua",
    description="Replace numerals with numbers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DanATW/numberize.git",
    project_urls={
        "Bug Tracker": "https://github.com/DanATW/numberize/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='text replace numerals words numbers morphology',
    packages=setuptools.find_packages(include=["numberize"]),
    python_requires=">=3.6",
    install_requires=['pymorphy2', 'pymorphy2-dicts-uk', 'nltk'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==6.2.4'],
    test_suite='test'
)
